# Copyright 2014-2015 Novartis Institutes for Biomedical Research

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os, subprocess
import warnings
import abc
import argparse
import tempfile
import logging
import re
import csv
logger = logging.getLogger(__name__)

from railroadtracks import core, environment, unifex
from railroadtracks.model.files import (BAMFile,
                                        BAMFileSortedByID,
                                        SamtoolsSorterByID,
                                        GFFFile,
                                        CSVFile,
                                        FilePattern,
                                        FASTQPossiblyGzipCompressed)

from railroadtracks.model.aligners import SailfishIndex

from railroadtracks.model.diffexp import RSOURCES_DIR

class ACTIVITY(core.Enum):
    """ Activities that can be performed by the different steps modeled.
    (note: steps can combine several activities - the most obvious example is
    existing monolithic pipelines) """
    __order__ = 'QUANTIFY'
    QUANTIFY = 'Quantify'

# -- assetsquantifier-begin
class AssetsQuantifier(core.AssetsStep):
    Source = core.assetfactory('Source', [core.AssetAttr('alignedreads', BAMFile, ''),
                                          core.AssetAttr('annotationfile', GFFFile, '')])
    Target = core.assetfactory('Target', [core.AssetAttr('counts', CSVFile, '')])
# -- assetsquantifier-end

# -- quantifierabstract-begin
class QuantifierAbstract(core.StepAbstract):
    __metaclass__ = abc.ABCMeta
    Assets = AssetsQuantifier
    activities = (ACTIVITY.QUANTIFY, )
# -- quantifierabstract-end

# -- htseqcount-begin
class HTSeqCount(QuantifierAbstract):
        
    _name = 'htseqcount'
    _default_execpath = 'htseq-count'
    _separator = '\t'
    # set of parameters to get htseq-count to work with references such
    # as bacteria or viruses (stored as a class attribute for convenience)
    _noexons_parameters = ('--type=CDS', '--idattr=db_xref', '--stranded=no')

    def __init__(self, executable=None):
        if executable is None:
            executable = type(self)._default_execpath
        self._execpath = executable
        self._version = None

    @property
    def version(self):
        if self._version is None:
            cmd = [self._execpath, '-h']
            try:
                logger.debug(subprocess.list2cmdline(cmd))
                res = subprocess.check_output(cmd)
            except OSError as ose:
                raise unifex.UnifexError("""Command: %s
                %s""" % (' '.join(cmd), ose))

            m = re.match('^.*( version )?([^ \n]+)\.$', res.split('\n')[-2])
            if m is None:
                raise RuntimeError('Could not find the version number.')
            self._version = m.groups()[1]
        return self._version

    def run(self, assets, parameters = tuple()):
        # FIXME: shouldn't strandedness be a better part of the model ?
        source = assets.source
        sortedbam = source.alignedreads
        if not isinstance(source.alignedreads, BAMFileSortedByID):
            # htseq-count requires sorted entries
            warnings.warn(("The source asset '%s' should ideally be sorted by read IDs. " +\
                           "We are sorting the file; use explicitly a '%s' rather than a '%s' "+\
                           "for better performances, as well as for reproducibility issues "+\
                           "(the sorting will use whatever 'samtools` is first found in the PATH)") \
                          % ("alignedreads", BAMFileSortedByID.__name__, BAMFile.__name__))
            output_dir = os.path.dirname(assets.target.counts.name)
            # temp file name for the sorted output
            sortedbam_fh = tempfile.NamedTemporaryFile(dir=output_dir, suffix=".bam", delete=False)
            # (cleaning temp files handled by Python, except sortedsam)
            # -- sort
            sorter = SamtoolsSorterByID()
            sorter_assets = sorter.Assets(sorter.Assets.Source(source.alignedreads),
                                          sorter.Assets.Target(BAMFile(sortedbam_fh.name)))
            sorter.run(sorter_assets)
            # sanity check:
            if os.stat(sorter_assets.target.sortedbam.name).st_size == 0:
                warnings.warn('The sorted BAM file is empty.')
            sortedbam = sorter_assets.target.sortedbam
        else:
            sortedbam_fh = None

        # BAM to SAM
        cmd_bam2sam = ['samtools', 'view', sortedbam.name]

        # build command line
        cmd_count = [self._execpath, ]
        cmd_count.extend(parameters)
        cmd_count.extend(['-', source.annotationfile.name])
        cmd = cmd_bam2sam + ['|', ] + cmd_count

        logger.debug(subprocess.list2cmdline(cmd))
        with open(os.devnull, "w") as fnull, \
             open(assets.target.counts.name, 'w') as output_fh:
            csv_w = csv.writer(output_fh)
            # HTSeq-count does not use column names in its output, unfortunately,
            # so we correct that
            csv_w.writerow(['ID','count'])
            with core.Popen(cmd_bam2sam, 
                            stdout=subprocess.PIPE,
                            stderr=fnull) as p_bam2sam:
                with p_bam2sam.stdout, core.Popen(cmd_count, 
                                                  stdin = p_bam2sam.stdout,
                                                  stdout = subprocess.PIPE,
                                                  stderr = subprocess.PIPE) as p_htseq:
                    p_bam2sam.stdout.close()
                    # read the output of HTSeq line-per-line
                    csv_r = csv.reader(p_htseq.stdout, delimiter='\t')
                    for row in csv_r:
                        csv_w.writerow(row)
                        p_htseq.stdout.flush()
                    stdout, stderr = p_htseq.communicate()
                if p_htseq.returncode != 0:
                    if sortedbam_fh is not None:
                        os.unlink(sortedbam_fh.name)
                    logger.error(subprocess.list2cmdline(cmd), extra={'stderr': stderr})
                    raise subprocess.CalledProcessError(p_htseq.returncode, cmd, None)
        if sortedbam_fh is not None:
            os.unlink(sortedbam_fh.name)
        return (cmd, p_htseq.returncode)
# -- htseqcount-end

class SailfishQuant(QuantifierAbstract):
        
    _name = 'sailfish-quant'
    _default_execpath = 'sailfish'

    LIBRARY_SE_DS = "'T=SE:S=U'"
    LIBRARY_PE = "'T=PE:O=><:S=SA'"
    TRANSCRIPT_COLNAME = '# Transcript'

    parser = argparse.ArgumentParser(_name)
    parser.add_argument('--libtype',
                        required=True,
                        help='Library type for Sailfish (see the documentation for Sailfish for details).'
                        'Canned values for that parameters are available as class attribute (LIBRARY_*).')
    parser.add_argument('--quantfile',
                        default='quant_bias_corrected.sf',
                        help='Quantification file produced by sailfish')
    parser.add_argument('--count-colname',
                        dest = 'count_colname',
                        default='EstimatedNumReads',
                        help='Column to use to report counts')

    class Assets(AssetsQuantifier):
        Source = core.assetfactory('Source', 
                                   [SailfishIndex.Assets.Target.getassetattr('indexfilepattern'),
                                    core.AssetAttr('read1', FASTQPossiblyGzipCompressed, ''),
                                    core.AssetAttr('read2', FASTQPossiblyGzipCompressed, '', allownone=True)])
        # 'counts' is coming from QuantifierAbstract
        Target = core.assetfactory('Target', [core.AssetAttr('counts', CSVFile, ''),
                                              core.AssetAttr('output_dir', FilePattern, '')])
                                              


    def __init__(self, executable=None):
        if type(self) == SailfishQuant:
            warnings.warn('Model SailfishQuant is deprecated. Please use SalmonAlignQuant.')

        if executable is None:
            executable = type(self)._default_execpath
        self._execpath = executable
        self._version = None

    @property
    def version(self):
        if self._version is None:
            cmd = [self._execpath, '--version']
            try:
                logger.debug(subprocess.list2cmdline(cmd))
                res = subprocess.check_output(cmd, stderr=subprocess.STDOUT)
            except OSError as ose:
                raise unifex.UnifexError("""Command: %s
                %s""" % (' '.join(cmd), ose))

            m = re.match('^version : (.+)$', res)
            if m is None:
                raise RuntimeError('Could not find the version number.')
            self._version = m.groups()[0]
        return self._version


    @staticmethod
    def _is_libtype_pe(libtype):
        m = re.match("^'T=(SE|PE).+'", libtype)
        if m is None:
            return None
        elif m.groups()[0] == 'PE':
            return True
        else:
            return False

    def run(self, assets, parameters = ('--libtype', LIBRARY_SE_DS, )):
        options, unknown = self.parser.parse_known_args(parameters)
        source = assets.source
        indexfiles = [name for cls, name in source.indexfilepattern.iterlistfiles()]
        if len(indexfiles) == 0:
            raise ValueError("No index files in %s" % source.indexfilepattern)

        # different strings for the command line
        if assets.source.read1 is None or not isinstance(source.read1, core.SavedEntityAbstract):
            raise ValueError("Incorrect value %s for read1" % source.read1)

        is_pe = self._is_libtype_pe(options.libtype)
        if is_pe is None:
            warnings.warn('Could not extract the library type from "%s".' % options.libtype)
        elif assets.source.read2 is None and is_pe:
            raise ValueError('The library type "%s" is for paired-ends but the "read2" is not defined in the assets.')

        if assets.source.read2 is None:
            # single reads
            cmd_sub = ('-r', source.read1.name)
        else:
            cmd_sub = ('-1', source.read1.name, 
                       '-2', source.read2.name)

        cmd_count = [self._execpath, 'quant']
        #cmd_count.extend(options)
        #cmd_count.extend(unknown)
        cmd_count.extend(parameters)
        cmd_count.extend(('--index', source.indexfilepattern.name))
        if os.path.exists(assets.target.output_dir.name):
            raise ValueError('%s will fail if the output directory "%s" is already present (and it is)' % (self._name, assets.target.counts.name))
        cmd_count.extend(('--out', assets.target.output_dir.name))
        cmd_count.extend(cmd_sub)
        #FIXME: why can't I make it work without list2cmdline() and shell=True
        with open(os.devnull, 'w') as fnull:
            returncode = subprocess.check_call(subprocess.list2cmdline(cmd_count), shell=True,
                                               stdout = fnull, stderr = fnull)

        # The model is trying to make quantifiers compatible with one an other.
        # We build the target asset `counts`
        # NOTE: We experienced that the Sailfish family could duplicate entries IDs
        #       (column TRANSCRIPT_COLNAME) while the numerical values differ (and be very close to zero)
        transcript_set = set()
        with open(assets.target.counts.name, 'w') as fh_out,\
             open(os.path.join(assets.target.output_dir.name, options.quantfile)) as fh_in:
            csv_r = csv.reader(fh_in, delimiter='\t')
            csv_w = csv.writer(fh_out)
            for row in csv_r:
                if (len(row) == 0) or row[0].startswith('#'):
                    header = row
                    continue
                break
            csv_w.writerow(['ID', 'count'])
            col_id_i = header.index(self.TRANSCRIPT_COLNAME)
            col_counts_i = header.index(options.count_colname)
            if header != row:
                transcript_id = row[col_id_i]
                transcript_set.add(transcript_id)
                csv_w.writerow((transcript_id, row[col_counts_i]))
            for row in csv_r:
                transcript_id = row[col_id_i]
                if transcript_id in transcript_set:
                    warnings.warn("The ID %s is not unique." % transcript_id)
                    continue
                transcript_set.add(transcript_id)
                csv_w.writerow((transcript_id, row[col_counts_i]))
        return (cmd_count, returncode)

class SalmonAlignQuant(SailfishQuant):
    _name = 'salmon-alignquant'
    _default_execpath = 'salmon'
    LIBRARY_SE_DS = None
    LIBRARY_PE = "IU"
    TRANSCRIPT_COLNAME = "# Name"
    parser = argparse.ArgumentParser(_name)
    parser.add_argument('--libType',
                        dest='libtype',
                        required=True,
                        help='Library type for Salmon (see the documentation for Sal,on for details).'
                        'Canned values for that parameters are available as class attribute (LIBRARY_*).')
    parser.add_argument('--quantfile',
                        default='quant.sf',
                        help='Quantification file produced by salmon')
    parser.add_argument('--count-colname',
                        dest = 'count_colname',
                        default='TPM',
                        help='Column to use to report counts')


class FeatureCount(QuantifierAbstract):
    _name = 'featurecount'
    _default_execpath = 'R'
    _rscript_name = 'featurecount.R'
    _rpackagename = 'Rsubread'

    # parser for additional parameters
    #parser = argparse.ArgumentParser('unifex')
    parser = argparse.ArgumentParser(_name)
    # parser.add_argument('action', nargs='?',
    #                     default = 'run')
    # parser.add_argument('model', nargs='?',
    #                     default = _name)
    # parser.add_argument('executable', nargs='?',
    #                     default = _default_execpath)
    parser.add_argument('--paired-ends',
                        dest = 'ispairedend',
                        default = False,
                        action='store_true')
    parser.add_argument('--strand-specific',
                        dest = 'strandspecific',
                        type = int,
                        default = 0,
                        choices = (0,1,2))
    parser.add_argument('--gtf-featuretype',
                        dest = 'gtf_featuretype',
                        default = 'exon',
                        help = "Passed to R's 'Rsubread::featureCounts()' as named parameter 'GTF.featureType' (default: %(default)s).")
    parser.add_argument('--gtf-attrtype',
                        dest = 'gtf_attrtype',
                        default = 'gene_id',
                        help = "Passed to R's 'Rsubread::featureCounts()' as named parameter 'GTF.attrType' (default: %(default)s).")

    # set of parameters to get htseq-count to work with references such
    # as bacteria or viruses (stored as a class attribute for convenience)
    _noexons_parameters = ('--gtf-featuretype=CDS', '--gtf-attrtype=db_xref')
    _pe_parameters = ('--paired-ends', )

    def __init__(self, executable=None):
        """
        :param executable: the executable is R. If None, the class-level
        attribute :attr:`_default_execpath` will be used.
        :type executable: a :class:`str` or a :class:`environment.R`
        """
        if executable is None:
            executable = type(self)._default_execpath
        if not isinstance(executable, environment.R):
            executable = environment.R(executable)
        self.r = executable
        rsource_template = os.path.join(RSOURCES_DIR, 
                                        self._rscript_name)
        assert os.path.isfile(rsource_template), \
            'The needed R script "%s" is not a file' % rsource_template
        self.rsource_template = rsource_template
        self._run_cmd = None
        self._version = None

    _execpath = property(lambda x: x.r.path, None, None)

    @property
    def version(self):
        if self._version is None:
            self._version = self.r.packageversion(self._rpackagename)
        return self._version

    def run(self, assets, parameters = tuple(), magicvariable = 'railroadtracks_import'):
        """ 
        :param assets: 
        :type assests: instance of :class:`core.AssetsStep` (or child class)
        :param targets:
        :type parameters: :class:`tuple`
        """
        assert isinstance(assets, AssetsQuantifier), "The parameter 'assets' must be an %s" % AssetsQuantifier.__name__

        options = self.parser.parse_args(parameters)
        var_in = {'alignedreads_fn': assets.source.alignedreads.name,
                  'annotation_fn': assets.source.annotationfile.name,
                  'results_fn': assets.target.counts.name,
                  'ispairedend': options.ispairedend,
                  'strandspecific': options.strandspecific,
                  'gtf_featuretype': options.gtf_featuretype,
                  'gtf_attrtype': options.gtf_attrtype}

        with open(self.rsource_template) as template:
            code = template.read()

        code = os.linesep.join([code, 'run(%s)' % magicvariable])
        returncode = self.r.run_snippet(code, var_in=var_in)
        # FIXME: should return the unified execution command line
        uei = core.UnifiedExecInfo(self.r.path, self._name, 
                                   assets.source, assets.target, parameters, 
                                   None, None # logging_file and logging_level
        )
        cmd = unifex._cmdfromuei(uei)
        return (cmd, returncode)




# Copyright 2014-2015 Novartis Institutes for Biomedical Research

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import os, subprocess, sys, shutil
from railroadtracks import hortator, unifex
from collections import namedtuple, OrderedDict, deque, Counter, defaultdict
import itertools

_TASK_DONE = hortator._TASK_DONE
_TASK_TODO = hortator._TASK_TODO
_TASK_INPROGRESS = hortator._TASK_INPROGRESS
_TASK_FAILED = hortator._TASK_FAILED

DbID = namedtuple('DbID', 'id')

def command_line(project, stepconcrete_id, stepobj, assets, 
                 parameters=(),
                 python_executable=sys.executable,
                 logging_file=None,
                 logging_level=None):
    """ 
    :param stepobj: Step object
    :type stepobj: instance of :class:`StepAbstract`
    :param assets: assets
    :type assets: instance of :class:`AssetsStep`
    :rtype: a named :class:`tuple` with the arguments for a command line call
    """
    import sys
    executable = stepobj._execpath
    model = type(stepobj)._name
    # FIXME: duplicated elsewhere in the same module - make a function
    cmd = [python_executable, '-m', 'railroadtracks.unifex', 'run', model]
    if executable is not None:
        cmd.append(executable)
    cmd.extend(['--module', stepobj.__class__.__module__])
    cmd.append('-s')
    field_items = list()
    for field,src in zip(assets.source._fields, assets.source):
        if src is None:
            continue
        for item in src:
            field_items.append((field,item))
    cmd.extend('%s=%s' % x for x in field_items)
    cmd.append('-t')
    cmd.extend('%s=%s' % (field,item) for field,src in zip(assets.target._fields, assets.target) for item in src)
    if logging_file is not None:
        cmd.append('--logging-file=%s' % logging_file)
    if logging_level is not None:
        cmd.append('--logging-level=%s' % logging_level)
    if len(parameters) > 0:
        cmd.append('-p')
        cmd.extend(' %s' % x for x in parameters)
    return cmd


#FIXME: rename this function
def get_steplist_up(project, stored_entities):
    def func_stored_entity(stored_entity):
        pass
    def func_stored_sequence(stored_entity):
        pass
    def func_step_concrete_factory(res):
        def f(step_concrete):
            res.append(step_concrete)
        return f
    def func_storedentity_stepconcrete(stored_entity, step_concrete):
        pass
    def func_stepconcrete_storedentity(step_concrete, stored_entity):
        pass
    all_stepconcrete = list()
    for stored_entity in stored_entities:
        res = list()
        project.cache.provenancewalk_storedentity(stored_entity,
                                                  func_stored_entity,
                                                  func_stored_sequence,
                                                  func_step_concrete_factory(res),
                                                  func_storedentity_stepconcrete,
                                                  func_stepconcrete_storedentity)
        all_stepconcrete.append(res)
    return all_stepconcrete



class Task(object):
    """ A task. """
    def __init__(self, project, call, task_id):
        self._call = call
        self._project = project
        if hasattr(task_id, 'id'):
            self._task_id = task_id.id
            self.is_new = task_id.new
        else:
            self._task_id = task_id
            self.is_new = None

    @property
    def activities(self):
        return self.call.step.activities

    @property
    def assets(self):
        return self.call.assets
    
    @property
    def call(self):
        """ A :class:`railroadtracks.unifex.Call` object. """
        return self._call
    @property
    def task_id(self):
        """ ID of the task in the project. """
        return self._task_id
    @property
    def info(self):
        """ Status information for the task. """
        # wrap into a DbID
        taskid = DbID(self._task_id)
        step_info = self._project.persistent_graph.step_concrete_info(taskid)
        return step_info
    @property
    def project(self):
        """ Project in which the task is defined. """
        return self._project
    @property
    def dirname(self):
        """ Directory in which the targets of the task are saved. """
        dirname = os.path.join(self.project.wd,
                               self.project.cache.stepconcrete_dirname(DbID(self.task_id)))
        return dirname

    def reset(self, ignore_errors=False):
        """ 'Reset' a task by setting the status to "TO DO" and erasing all result files in the task directory."""
        dirname = self.dirname
        self.status = _TASK_TODO
        for entry in os.listdir(dirname):
            if os.path.isdir(entry):
                shutil.rmtree(os.path.join(dirname, entry), ignore_errors=ignore_errors)
            else:
                os.unlink(os.path.join(dirname, entry))
        
    def unifex_cmd(self, python_executable=sys.executable,
                   logging_file=None,
                   logging_level=None):
        """ Parsed command to run the task. 

        To make a string to copy/paste into a shell script,
        use :func:`subprocess.list2cmdline`.

        """
        cmd = command_line(self._project, 
                           self._task_id, #FIXME: would work with _task_id ?
                           self._call.step, 
                           self._call.assets, 
                           parameters=self._call.parameters,
                           python_executable=python_executable,
                           logging_file=logging_file,
                           logging_level=logging_level)

        return cmd

    def unifex_cmdline(self, python_executable=sys.executable):
        """ 
        Command line as could be run in a shell.
        ( This is a wrapper around subprocess.list2cmdline(self.unifex_cmd()) )
        """
        return subprocess.list2cmdline(self.unifex_cmd(python_executable=python_executable))


    def execute(self):
        """ Execute the task. Note that the status of the task known to the project is not updated. """
        cmd,returncode = self._call.execute()
        if returncode != 0:
            raise unifex.UnifexError('The execution not successful with\n%s\nCheck/activate the logs to trace what is happening.' % str(self))

    def __str__(self):
        res = (super(Task, self).__str__(),
               'task ID: %i' % self.task_id,
               str(self.info),
               str(self.call),
               '---')
        return os.linesep.join(res)


    def parent_tasks(self):
        """ Tasks the current task is depending on. """
        res = TaskSet()
        # wrap into a DbID
        taskid = DbID(self._task_id)
        # fetch the source assets
        entities = self.project.persistent_graph.get_srcassets(taskid)
        for et in entities:
            # for each source asset, look for the tasks that also have it as a target
            if hasattr(et, 'iter_storedentities'):
                for sub_et in et.iter_storedentities():
                    sc_id = self.project.persistent_graph.get_parenttask_of_storedentity(sub_et)
                    if sc_id is None:
                        # we reached a root node
                        continue
                    # only one task at the origin of a target
                    res.add(self.project.get_task(sc_id))
            else:
                sc_id = self.project.persistent_graph.get_parenttask_of_storedentity(et)
                if sc_id is None:
                    # we reached a root node
                    continue
                # only one task at the origin of a target
                res.add(self.project.get_task(sc_id))
        return res

    def primordial_tasks(self):
        """ Direct or indirect parent tasks with source assets that do not
        have parent tasks themselves.
        In other words, root nodes in the dependency graph connected to this task.

        This method is useful to identify the raw data results are derived from.
        """
        res = TaskSet()
        # start with the targets because the current task may also have 
        # primordial source assets
        targets = tuple(self.iter_targetassets())
        up = get_steplist_up(self.project, [x.id for x in targets])
        for tsk in (y for x in up for y in x):
            ancestor_task = self.project.get_task(tsk)
            # The source assets, that is the assets used by the task, can be extracted
            for src_asset in ancestor_task.iter_sourceassets():
                parent_task = src_asset.parenttask
                if parent_task is None:
                    # This source asset does not have a parent task in the project
                    # It is therefore considered a root asset in railroadtracks
                    # (the asset might originate from earlier steps, 
                    # but they are not stored in the project)
                    res.add(ancestor_task)
                    # The task is in, no need to iterate further
                    break
        return res

    def child_tasks(self):
        """ Tasks depending on the current task. """
        res = TaskSet()
        # wrap into a DbID
        taskid = DbID(self._task_id)
        # fetch the target assets
        entities = self.project.persistent_graph.get_targetassets(taskid)
        for et in entities:
            # for each source asset, look for the tasks that also have it as a source
            if hasattr(et, 'iter_storedentities'):
                for sub_et in et.iter_storedentities():
                    for sc_id in self.project.persistent_graph.get_targetstepconcrete(sub_et):
                        if sc_id is None:
                            # root node
                            continue
                        res.add(self.project.get_task(sc_id))
            else:
                for sc_id in self.project.persistent_graph.get_targetstepconcrete(et):
                    if sc_id is None:
                        # root node
                        continue
                    res.add(self.project.get_task(sc_id))
        return res

    def all_child_tasks(self):
        """ Get all the child tasks, direct or indirect. """
        res = TaskSet()
        nodes_visited = set()
        task_queue = deque()
        # get the tasks downstream
        for t in self.child_tasks():
            if t.task_id in nodes_visited:
                continue
            task_queue.append(t)
        while len(task_queue) > 0:
            task = task_queue.pop()
            if task in res:
                continue
            nodes_visited.add(task.task_id)
            res.add(task)
            # get the tasks downstream
            for t in task.child_tasks():
                if t.task_id in nodes_visited:
                    continue
                task_queue.append(t)
        return res

    @property
    def status(self):
        """ Get/Set the status of the task in the project.
        The value must be a valid task status.
        """
        # wrap into a DbID
        taskid = DbID(self._task_id)
        res = self._project.persistent_graph.step_concrete_status(taskid)
        #FIXME: this will likely change at the lower level, the assertion will
        #       act as watcher
        assert(isinstance(res, list))
        assert(len(res) == 1)
        assert(len(res[0]) == 2)
        return res[0][1]
        
    @status.setter
    def status(self, value):
        # wrap into a DbID
        taskid = DbID(self._task_id)
        self._project.persistent_graph.step_concrete_state(taskid,
                                                           hortator._TASK_STATUS_LIST[value])

    @property
    def time_points(self):
        # wrap into a DbID
        taskid = DbID(self._task_id)
        return self._project.persistent_graph.task_time_points(taskid)

    def iter_sourceassets(self):
        return self.project.iter_srcassets(self)

    def iter_targetassets(self):
        return self.project.iter_targetassets(self)

    
class TaskSet(object):
    """ Ordered set of tasks.
    This class can be used for independent tasks for which the order of completion does not matter,
    and provide a container for tasks that can run in parallel.
    """
    _d = None
    _s = None
    def __init__(self, project=None, iterable=tuple()):
        self._d = OrderedDict()
        self._s = set()
        self._project = project
        for item in iterable:
            self.add(item)

    def _ensure_task_project(self, task):
        if self._project is None:
            self._project = task._project
        elif task._project != self._project:
            raise ValueError('Mismatching project.')
    def _ensure_taskset_project(self, taskset):
        if taskset._project is None:
            if len(taskset) > 0:
                raise Exception('We should not be here.')
            else:
                return
        if self._project is None:
            self._project = taskset._project
        elif taskset._project != self._project:
            raise ValueError('Mismatching project.')
    def __iter__(self):
        return iter(self._d.values())

    def iter_chunks(self, maxsize):
        """ Return an iterator of TaskSet of maximum size `maxsize`. """
        it = iter(self)
        l = len(self)
        for chunk_i in range(0, l, maxsize):
            if (chunk_i+maxsize) > l:
                chunksize = l - chunk_i
            else:
                chunksize = maxsize
            yield TaskSet(iterable=itertools.islice(it, 0, chunksize))
            
    def __contains__(self, task):
        self._ensure_task_project(task)
        return task.task_id in self._d
    def __len__(self):
        return len(self._d)
    def __str__(self):
        res = [super(TaskSet, self).__str__(),
               '  with %i element(s) of respective status:' % len(self)]
        res.extend('  - %s: %i' % (k, v) for k,v in self.status().items())
        return '\n'.join(res)
    def add(self, task):
        """ Add a task. """
        assert isinstance(task, Task), 'Element should be in an instance of Task.'
        self._ensure_task_project(task)
        self._s.add(task.task_id)
        self._d[task.task_id] = task
    def remove(self, task):
        """ remove a task. """
        assert isinstance(task, Task), 'Element should be in an instance of Task.'
        self._ensure_task_project(task)
        self._s.remove(task.task_id)
        del(self._d[task.task_id])
    def union(self, taskset):
        assert isinstance(taskset, TaskSet), "'taskset' should be a instance of TaskSet."
        self._ensure_taskset_project(taskset)
        taskids = self._s.union(taskset._s)
        u = TaskSet()
        for task_id in taskids:
            try:
                task = self._d[task_id]
            except KeyError:
                task = taskset._d[task_id]
            u.add(task)
        return u
    def update(self, taskset):
        assert isinstance(taskset, TaskSet), "'taskset' should be a instance of TaskSet."
        self._ensure_taskset_project(taskset)
        self._s.update(taskset._s)
        for task_id in taskset._s:
            self._d.update(taskset._d)

    def intersection(self, taskset):
        assert isinstance(taskset, TaskSet), "'taskset' should be a instance of TaskSet."
        self._ensure_taskset_project(taskset)
        taskids = self._s.intersection(taskset._s)
        i = TaskSet()
        for task_id in taskids:
            task = self._d[task_id]
            i.add(task)
        return i
    def difference(self, taskset):
        assert isinstance(taskset, TaskSet), "'taskset' should be a instance of TaskSet."
        self._ensure_taskset_project(taskset)
        taskids = self._s.difference(taskset._s)
        i = TaskSet()
        for task_id in taskids:
            task = self._d[task_id]
            i.add(task)
        return i
    def isdisjoint(self, taskset):
        assert isinstance(taskset, TaskSet), "'taskset' should be a instance of TaskSet."
        self._ensure_taskset_project(taskset)
        res = self._s.isdisjoint(taskset._s)
        return res
    def issubset(self, taskset):
        assert isinstance(taskset, TaskSet), "'taskset' should be a instance of TaskSet."
        self._ensure_taskset_project(taskset)
        res = self._s.issubset(taskset._s)
        return res
    def issuperset(self, taskset):
        assert isinstance(taskset, TaskSet), "'taskset' should be a instance of TaskSet."
        self._ensure_taskset_project(taskset)
        res = self._s.issuperset(taskset._s)
        return res
    def values(self):
        """ Return the tasks as a tuple (Python 2) or an iterator (Python 3). """
        return self._d.values()
    def status(self):
        """ Return a `Counter` of status labels. """
        ct = Counter(x.status for x in self._d.values())
        return ct
    def split_by_status(self):
        res = defaultdict(TaskSet)
        for task in self:
            #FIXME: simplify access to status
            res[task.status].add(task)
        return res
    def filter_on_status(self, status):
        res = TaskSet()
        for task in self:
            #FIXME: simplify access to status
            if task.status == status:
                res.add(task)
        return res
    def filter_on_parent_status(self, status):
        res = TaskSet()
        for task in self:
            ok = True
            for parent_task in task.parent_tasks():
                if parent_task.status != status:
                    ok = False
                    break
            if ok:
                res.add(task)
        return res




# Copyright 2014-2015 Novartis Institutes for Biomedical Research

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
import tempfile
from railroadtracks import environment

has_R = environment.Executable.ispresent('R')

class EnvironmentTestCase(unittest.TestCase):

    def test_executable(self):
        ls_exec = environment.Executable('ls')
        #FIXME: technically no test here

    @unittest.skipIf(not has_R,
                     'R is missing')
    def test_R(self):
        r_exec = environment.R('R')
        r_version = r_exec.version
        # missing package
        self.assertRaises(ValueError, r_exec.packageversion, 'foobarbaz')

        version = r_exec.packageversion('stats')
        self.assertTrue(r_version.startswith(version))

    @unittest.skipIf(not (has_R and \
                          environment.R('R').packageversion_or_none('rjson') is not None),
                     'R and its package "rjson" must be present.')
    def test_R_run_snippet(self):
        r_exec = environment.R('R')
        # run a snippet
        magicvariable = 'railroadtracks_import'
        code = """        
        run <- function(p) {
          res <- 1 + p$x
          # output file
          fh_out <- file(p$output_fn, "w")
          write(res, file = fh_out)
          flush(fh_out)
        }
        run(%s)
        """ % magicvariable
        with tempfile.NamedTemporaryFile() as fh_out:
            var_in = {'x': 3,
                      'output_fn': fh_out.name}
            r_exec.run_snippet(code, var_in)
            res = fh_out.read()
        self.assertEqual(b'4', res.rstrip())
